#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# This file is part of the dodoo-migrator (R) project.
# Copyright (c) 2018 Camptocamp SA and XOE Corp. SAS
# Authors: Guewen Baconnier, Leonardo Pistone, David Arnold, et al.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, see <http://www.gnu.org/licenses/>.
#

import ast
import os
import subprocess

import pytest
from click.testing import CliRunner
from dodoo import odoo

from dodoo_migrator.cli import migrate

HERE = os.path.dirname(__file__)
DATADIR = os.path.join(HERE, "data/test_migrator/")
MIG_TABLE = "dodoo_migrator"
MANIFEST_NAMES = ("__manifest__.py", "__openerp__.py")


def _exec_query(dbname, query):
    process = subprocess.Popen(
        ["psql", "-d", dbname, "-t", "-c", query], stdout=subprocess.PIPE
    )
    out, _ = process.communicate()
    return out


def test_mig_sorting(odoodb, odoocfg):
    """ Test if migrations are sorted properly """

    result = CliRunner().invoke(
        migrate,
        [
            "-d",
            odoodb,
            "-c",
            str(odoocfg),
            "--file",
            DATADIR + ".mig-0.0.x-sorting.yaml",
        ],
    )
    assert result.exit_code == 0
    result = _exec_query(odoodb, "SELECT number FROM {}".format(MIG_TABLE))
    # Assert that log entries are create in the right order.
    assert result == b" 0.0.1\n 0.0.2\n 0.0.3\n\n"


def test_migrator_operations(odoodb, odoocfg):
    """ Test all migrator operations """

    # Test install, upgrade
    result = CliRunner().invoke(
        migrate,
        [
            "-d",
            odoodb,
            "-c",
            str(odoocfg),
            "--file",
            DATADIR + ".mig-0.1.1-install.yaml",
        ],
    )
    assert result.exit_code == 0

    result = _exec_query(
        odoodb, "SELECT state FROM ir_module_module WHERE name='board'"
    )
    # Assert that mail is installed.
    assert result == b" installed\n\n"

    # Test uninstall
    result = CliRunner().invoke(
        migrate,
        [
            "-d",
            odoodb,
            "-c",
            str(odoocfg),
            "--file",
            DATADIR + ".mig-0.1.2-uninstall.yaml",
        ],
    )
    assert result.exit_code == 0
    result = _exec_query(
        odoodb, "SELECT state FROM ir_module_module WHERE name='board'"
    )
    # Assert that mail is uninstalled.
    assert result == b" uninstalled\n\n"

    if odoo.release.version_info[0] > 10:
        # Test remove (via odoo.migration package)
        result = CliRunner().invoke(
            migrate,
            [
                "-d",
                odoodb,
                "-c",
                str(odoocfg),
                "--file",
                DATADIR + ".mig-0.1.3-remove.yaml",
            ],
        )
        assert result.exit_code == 0
        result = _exec_query(
            odoodb, "SELECT 1 FROM ir_module_module WHERE name='board'"
        )
        # Assert board has been removed from module index.
        assert result == b"\n"


def test_database_advisory_lock(odoodb, odoocfg):
    """ Test that no two migrations can accidentially run in parallel """
    pass


def test_migr_folder_overlay(odoodb, odoocfg):
    """ Test if migration folder overlay is workging correctly and
    upgrade scrips"""
    odoo.tools.config["stop_after_init"] = False
    found = False
    for manifest in MANIFEST_NAMES:
        web_manifest_path = os.path.join(
            odoo.__path__[0], "..", "addons", "web", manifest
        )
        if os.path.isfile(web_manifest_path):
            found = True
            break
    if not found:
        raise

    with open(web_manifest_path, "r") as f:
        old_str = f.read()
        info = ast.literal_eval(old_str)
        info["version"] = "5.0"

    with open(web_manifest_path, "w") as f:
        f.write(str(info))

    with open(web_manifest_path, "r") as f:
        test_str = f.read()

    assert "5.0" in test_str

    result = CliRunner().invoke(
        migrate,
        [
            "-m",
            DATADIR + "overlay/",
            "-d",
            odoodb,
            "-c",
            str(odoocfg),
            "--file",
            DATADIR + ".mig-0.1.4-upgrade.yaml",
        ],
    )
    with open(web_manifest_path, "w") as f:
        f.write(old_str)

    assert result.exit_code == 0

    result_pre = _exec_query(
        odoodb, "SELECT 1 FROM dodoo_test_migrations WHERE name='pre'"
    )
    # Assert pre entry has been written.
    assert result_pre == b"        1\n\n"

    result_post = _exec_query(
        odoodb, "SELECT 1 FROM dodoo_test_migrations WHERE name='post'"
    )
    # Assert post entry has been written.
    assert result_post == b"        1\n\n"

    if odoo.release.version_info[0] <= 8:
        return

    result_end = _exec_query(
        odoodb, "SELECT 1 FROM dodoo_test_migrations WHERE name='end'"
    )
    # Assert end entry has been written.
    assert result_end == b"        1\n\n"


def test_pre_and_post_scripts(odoodb, odoocfg):
    """ Test pre and post scripts are executing properly. """
    # Test install, upgrade
    result = CliRunner().invoke(
        migrate,
        [
            "-d",
            odoodb,
            "-c",
            str(odoocfg),
            "--file",
            DATADIR + ".mig-0.2.0-pre-post-scripts.yaml",
        ],
    )
    assert result.exit_code == 0
    result_pre = _exec_query(
        odoodb,
        """SELECT udt_name
                     FROM information_schema.columns
                    WHERE table_name = 'res_partner'
                      AND column_name = 'pre_script_column_wo_mig'""",
    )
    # Assert end entry has been written.
    assert result_pre == b" varchar\n\n"
    result_pre = _exec_query(
        odoodb,
        """SELECT udt_name
                     FROM information_schema.columns
                    WHERE table_name = 'res_partner'
                      AND column_name = 'post_script_column_wo_mig'""",
    )
    # Assert end entry has been written.
    assert result_pre == b" varchar\n\n"


def test_namespaced_mig_module(odoodb, odoocfg):
    """ Test if name-spaced odoo module is working properly """
    if odoo.release.version_info[0] <= 10:
        pytest.skip("version <= 10 does not have the namespaced migration module.")
    # Test install, upgrade
    result = CliRunner().invoke(
        migrate,
        [
            "-d",
            odoodb,
            "-c",
            str(odoocfg),
            "--file",
            DATADIR + ".mig-0.3.0-pre-post-scripts-using-mig-module.yaml",
        ],
    )
    assert result.exit_code == 0
    result_pre = _exec_query(
        odoodb,
        """SELECT udt_name
                     FROM information_schema.columns
                    WHERE table_name = 'res_partner'
                      AND column_name = 'pre_script_column'""",
    )
    # Assert end entry has been written.
    assert result_pre == b" varchar\n\n"
    result_pre = _exec_query(
        odoodb,
        """SELECT udt_name
                     FROM information_schema.columns
                    WHERE table_name = 'res_partner'
                      AND column_name = 'post_script_column'""",
    )
    # Assert end entry has been written.
    assert result_pre == b" varchar\n\n"



import sys

from odoo import migration

migration.module_installed(cr, "base")  # noqa
# ALSO TOUCHES: modules_installed
migration.remove_module(cr, "iap")  # noqa
# ALSO TOUCHES: table_of_model
# ALSO TOUCHES: table_exists
# ALSO TOUCHES: remove_view
# ALSO TOUCHES: remove_record
# ALSO TOUCHES: remove_menus
# ALSO TOUCHES: remove_model
# ALSO TOUCHES: indirect_references
# ALSO TOUCHES: column_exists
# ALSO TOUCHES: column_type
# ALSO TOUCHES: table_exists
# ALSO TOUCHES: view_exists
# ALSO TOUCHES: remove_field
# ALSO TOUCHES: remove_refs
migration.rename_module(cr, "web_tour", "web_tours")  # noqa
migration.merge_module(cr, "web_tours", "web")  # noqa
migration.force_install_module(cr, "sale", deps=["mrp"])  # noqa
migration.force_install_module(cr, "sale", deps=None)  # noqa
migration.module_deps_diff(cr, "sale", ("stock",), ("stock",))  # noqa
# ALSO TOUCHES: new_module_dep
# ALSO TOUCHES: force_install_module
# ALSO TOUCHES: remove_module_deps

# TODO: Needs special preparation
# migration.new_module(cr,  module, deps=(), auto_install=False)  # noqa
# ALSO TOUCHES: new_module_dep
# ALSO TOUCHES: modules_installed

# NOTE: Cannot be readily tested as it analyzes the executin environment
# and infers a version from the file
# migration.force_migration_of_fresh_module(cr, "sale")  # noqa
migration.split_group(cr, "base.group_user", "base.group_system")  # noqa

migration.create_m2m(
    cr, "rel_tbl", "res_partner", "res_company", col1=None, col2=None
)  # noqa
migration.create_m2m(
    cr, "rel_tbl2", "res_partner", "res_company", col1="col1", col2="col2"
)  # noqa
migration.ensure_m2o_func_field_data(
    cr, "res_partner", "user_id", "res_company"
)  # noqa
# ALSO TOUCHES: column_exists
# ALSO TOUCHES: remove_column

# TODO: https://github.com/odoo/odoo/pull/26548#pullrequestreview-234117381
# migration.uniq_tags(cr, "res.partner.category", uniq_column="name", order="id")  # noqa
# ALSO TOUCHES: table_of_model
# ALSO TOUCHES: get_columns


migration.move_field_to_module(cr, "res.partner", "color", "base", "web")  # noqa
migration.rename_field(
    cr, "res.partner", "name", "name2", update_references=True
)  # noqa
# ALSO TOUCHES: table_of_model
# ALSO TOUCHES: table_exists
# ALSO TOUCHES: column_exists
# ALSO TOUCHES: update_field_references
migration.make_field_company_dependent(
    cr, "res.partner", "title", "many2one", target_model="res.partner.title"
)  # noqa
# ALSO TOUCHES: remove_column
# ALSO TOUCHES: drop_depending_views
# ALSO TOUCHES: get_depending_views

# TODO: We most probably destroyed the schema by now
# migration.recompute_fields(cr, model, fields, ids=None, logger=_logger, chunk_size=256)  # noqa

# TODO: check what it does
# migration.fix_wrong_m2o(cr, table, column, target, value=None)


migration.move_model(
    cr, "res.lang", "base", "web", move_data=True, delete=False
)  # noqa
migration.rename_model(
    cr, "res.currency.rate", "res.currency.rates", rename_table=True
)  # noqa
# ALSO TOUCHES: res_model_res_id

migration.replace_record_references(
    cr, ("res.partner", 2), (None, 1), replace_xmlid=True
)  # noqa
# ALSO TOUCHES: replace_record_references_batch
# ALSO TOUCHES: get_fk

migration.create_column(cr, "res_partner", "new_col", "varchar")  # noqa

# TODO: construe a test case
# migration.delete_unused(cr, table, xmlids, set_noupdate=True)  # noqa
# ALSO TOUCHES: force_noupdate

migration.get_index_on(cr, "res_currency", ["name"])  # noqa

migration.rename_xmlid(
    cr, "base.main_company", "base.main_company2", noupdate=None
)  # noqa
# Found
migration.ensure_xmlid_match_record(
    cr, "base.main_company2", "res.company", {"name": "My Company"}
)  # noqa
# Not Found (any more)
migration.ensure_xmlid_match_record(
    cr, "base.main_company", "res.company", {"name": "My Company"}
)  # noqa

cr.commit()

# We destroyed the database, prevent loading registry (which would fail)
sys.exit()



# -*- coding: utf-8 -*-
# Copyright 2018-2018 XOE Corp. SAS
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html)


SELECTION_GROUP_HIERARCHY = {"4": ["3", "2", "1"], "3": ["2", "1"], "2": ["1"], "1": []}


# fmt: off
KNOWN_FIELD_NON_SLOT_ATTRIBUTES = {
    'type'                : {'selection_group': '4'},        # type of the field (string)
    'relational'          : {'selection_group': '4'},        # whether the field is a relational one
    'translate'           : {'selection_group': '4'},        # whether the field is translated
}

KNOWN_FIELD_SLOTS = {

    # fields.MetaField
    'args'                : {'selection_group': '4'},       # the parameters given to __init__()
    '_attrs'              : {'selection_group': '4'},       # the field's non-slot attributes
    '_module'             : {'selection_group': '4'},       # the field's module name
    '_modules'            : {'selection_group': '4'},       # modules that define this field
    '_setup_done'         : {'selection_group': '4'},       # the field's setup state : None, 'base' or 'full'
    '_sequence'           : {'selection_group': '4'},       # absolute ordering of the field

    'automatic'           : {'selection_group': '4'},       # whether the field is automatically created ("magic" field)
    'inherited'           : {'selection_group': '4'},       # whether the field is inherited (_inherits)
    'inherited_field'     : {'selection_group': '4'},       # the corresponding inherited field

    'name'                : {'selection_group': '4'},       # name of the field
    'model_name'          : {'selection_group': '4'},       # name of the model of this field
    'comodel_name'        : {'selection_group': '4'},       # name of the model of values (if relational)

    'store'               : {'selection_group': '4'},       # whether the field is stored in database
    'index'               : {'selection_group': '4'},       # whether the field is indexed in database
    'manual'              : {'selection_group': '4'},       # whether the field is a custom field
    'copy'                : {'selection_group': '4'},       # whether the field is copied over by BaseModel.copy()
    'depends'             : {'selection_group': '4'},       # collection of field dependencies
    'recursive'           : {'selection_group': '4'},       # whether self depends on itself
    'compute'             : {'selection_group': '4'},       # compute(recs) computes field on recs
    'compute_sudo'        : {'selection_group': '4'},       # whether field should be recomputed as admin
    'inverse'             : {'selection_group': '4'},       # inverse(recs) inverses field on recs
    'search'              : {'selection_group': '4'},       # search(recs, operator, value) searches on self
    'related'             : {'selection_group': '4'},       # sequence of field names, for related fields
    'related_sudo'        : {'selection_group': '4'},       # whether related fields should be read as admin
    'company_dependent'   : {'selection_group': '4'},       # whether ``self`` is company-dependent (property field)
    'default'             : {'selection_group': '4'},       # default(recs) returns the default value

    'string'              : {'selection_group': '4'},       # field label
    'help'                : {'selection_group': '4'},       # field tooltip
    'readonly'            : {'selection_group': '4'},       # whether the field is readonly
    'required'            : {'selection_group': '4'},       # whether the field is required
    'states'              : {'selection_group': '4'},       # set readonly and required depending on state
    'groups'              : {'selection_group': '4'},       # csv list of group xml ids
    'change_default'      : {'selection_group': '4'},       # whether the field may trigger a "user-onchange"
    'deprecated'          : {'selection_group': '4'},       # whether the field is deprecated

    'related_field'       : {'selection_group': '4'},       # corresponding related field
    'group_operator'      : {'selection_group': '4'},       # operator for aggregating values
    'group_expand'        : {'selection_group': '4'},       # name of method to expand groups in read_group()
    'prefetch'            : {'selection_group': '4'},       # whether the field is prefetched
    'context_dependent'   : {'selection_group': '4'},       # whether the field's value depends on context

    # fields.Char
    'size'                : {'selection_group': '4'},       # maximum size of values (deprecated)
    'trim'                : {'selection_group': '4'},       # whether value is trimmed (only by web client)

    # fields.Float
    #  `_digits` slot is accessed through `digits` property function
    'digits'              : {'selection_group': '4'},        # digits argument passed to class initializer

    # fields.Monetary
    'currency_field'      : {'selection_group': '4'},

    # fields._String
    'translate'           : {'selection_group': '4'},       # whether the field is translated

    # fields.Html
    'sanitize'            : {'selection_group': '4'},       # whether value must be sanitized
    'sanitize_tags'       : {'selection_group': '4'},       # whether to sanitize tags (only a white list of attributes is accepted)
    'sanitize_attributes' : {'selection_group': '4'},       # whether to sanitize attributes (only a white list of attributes is accepted)
    'sanitize_style'      : {'selection_group': '4'},       # whether to sanitize style attributes
    'strip_style'         : {'selection_group': '4'},       # whether to strip style attributes (removed and therefore not sanitized)
    'strip_classes'       : {'selection_group': '4'},       # whether to strip classes attributes

    # fields.Binary
    'attachment'          : {'selection_group': '4'},       # whether value is stored in attachment

    # fields.Selection
    'selection'           : {'selection_group': '4'},       # [(value, string), ...], function or method name
    'validate'            : {'selection_group': '4'},       # whether validating upon write

    # fields._Relational
    'domain'              : {'selection_group': '4'},       # domain for searching values
    'context'             : {'selection_group': '4'},       # context for searching values
    'auto_join'           : {'selection_group': '4'},       # whether joins are generated upon search

    # fields.Many2one
    'ondelete'            : {'selection_group': '4'},       # what to do when value is deleted
    'delegate'            : {'selection_group': '4'},       # whether self implements delegation

    # fields._RelationalMulti
    'limit'               : {'selection_group': '4'},       # optional limit to use upon read

    # fields.Many2many
    'relation'            : {'selection_group': '4'},       # name of table
    'column1'             : {'selection_group': '4'},       # column of table referring to model
    'column2'             : {'selection_group': '4'},       # column of table referring to comodel

    # fields.One2many
    'inverse_name'        : {'selection_group': '4'},       # name of the inverse field
}
# fmt: on



# -*- coding: utf-8 -*-
# Copyright 2018-2018 XOE Corp. SAS
# License AGPL-3.0 or later (http://www.gnu.org/licenses/agpl.html)

import sys
from collections import OrderedDict

import pandas
import yaml

import odoo

from ..env import OdooAnalyzerEnvironment
from ..git import Git
from ._exceptions import ExtraColumnsException

DB_PREFIX = "dodoo-migrator-analyzer-temporary-branch-"

LIB_UPDATE_STR = "The analyzer package might need an update to handle this properly."


def ordered_load(stream, Loader=yaml.Loader, object_pairs_hook=OrderedDict):
    # This won't be necesary after python 3.7:
    # https://stackoverflow.com/a/21912744
    # "In python 3.7+, the insertion-order preservation nature of dict objects
    # has been declared to be an official part of the Python language spec"
    class OrderedLoader(Loader):
        pass

    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))

    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG, construct_mapping
    )
    return yaml.load(stream, OrderedLoader)


def get_dataframe_of_all_fields(registry):
    rows = []
    for model in registry.models:
        for field in model._fields:
            row = {}
            row["index"] = field.name.__str__()
            attributes = (
                # all slots of field type
                list(type(field)._slots.keys())
                # all non-slot attributes of field instance
                + list(field._attrs.keys())
            )
            for attr in attributes:
                # filter out dict containing non-slot attributes
                if attr == "_attrs":
                    continue
                if attr == "_modules":
                    # make set order relevant entropy in panda equality check
                    # _modules order representes the inverse mro
                    # list type also enables slicing
                    row[attr] = list(attr)
                    row["_origin_module"] = attr[-1]
                    continue
                row[attr] = field.__getattribute__(attr)

    return pandas.DataFrame.from_dict(rows, index="index")


class SchemaAnalyzer(object):
    """ Analyzes and compares Odoo registries of different code bases for their
    schema changes """

    def __init__(
        self,
        git_dir,
        old_branch,
        new_branch,
        exogenous_information_file=None,
        environment_manager=OdooAnalyzerEnvironment,
    ):
        # TODO: wire PYTHOPATH + GIT_DIR together to present a consistent source
        self.git_dir = git_dir
        sys.path = [git_dir] + sys.path
        self.old_branch = old_branch
        self.old_branch_db_name = DB_PREFIX + old_branch
        self.new_branch = new_branch
        self.new_branch_db_name = DB_PREFIX + new_branch
        (
            self.model_renames,
            self.field_renames,
            self.field_ignores,
        ) = self._parse_exogenous_information(exogenous_information_file)
        self.old_fields_df = pandas.DataFrame
        self.new_fields_df = pandas.DataFrame
        self.environment_manager = environment_manager

    def _load(self):
        """ Loads fields dataframes of both codebases"""
        with Git(git_dir=self.git_dir) as git:

            # Load analysis from old branch codebase
            git.checkout(self.old_branch)
            with self.environment_manager(self.old_branch_db_name) as env:
                self.old_fields_df = get_dataframe_of_all_fields(env.registry)

            # Load analysis from new branch codebase
            git.checkout(self.new_branch)
            with self.environment_manager(self.new_branch_db_name) as env:
                self.new_fields_df = get_dataframe_of_all_fields(env.registry)

    def _validate_exogenous_information(self, exogenous_information_dict):
        # No split check already done by yaml validation?
        # Check source and target are valid in model_renames and field_renames
        # Suggest module rename if all non-magic fields are moved
        # Check field_ignores does not cross with field_renames (ignores preced)
        return exogenous_information_dict

    def _normalize_exogenous_information(self, exogenous_information_dict):
        model_renames = exogenous_information_dict["model_renames"]
        field_renames = exogenous_information_dict["field_renames"]
        field_ignores = exogenous_information_dict["field_ignores"]
        return model_renames, field_renames, field_ignores

    def _parse_exogenous_information(self, exogenous_information_file):
        f_obj = open(exogenous_information_file, "rb")
        file_byte_content = f_obj.read()
        exogenous_information_dict = ordered_load(file_byte_content)
        exogenous_information_dict = self._validate_exogenous_information(
            exogenous_information_dict
        )
        model_renames, field_renames, field_ignores = self._normalize_exogenous_information(  # noqa: B950
            exogenous_information_dict
        )
        return model_renames, field_renames, field_ignores

    def _compare(self, selection_group=4):
        """ Compares panda dataframes for changes on the attributes
        included by the selection_group, also takes exogenous known information
        into account to reduce the diff """
        df1 = self.old_fields_df
        df2 = self.new_fields_df

        # Step 1:  Normalize indices and store delta
        # Step 1a: drop ignored fields on both, old and new dataframes
        df1 = df1.drop(self.field_ignores)
        df2 = df2.drop(self.field_ignores)
        # Step 1bi: inject exogenous relocation information (model renames)
        index = df1.index
        for source, target in self.model_renames.items():
            index = index.str.replace("^" + source, target, regex=True)
        df1.index = index
        # Step 1bii: inject exogenous relocation information (field renames)
        df1 = df1.rename(index=self.field_renames)
        rows = df1.index & df2.index
        df1_extra_rows = rows ^ df1.index
        df2_extra_rows = rows ^ df2.index

        # Step 2:  Normalize columns and store delta.
        # Step 2a: inject known column renames
        df1 = df1.rename(columns=dict(odoo.fields.RENAMED_ATTRS))
        columns = df1.columns & df2.columns
        df1_extra_columns = columns ^ df1.columns
        df2_extra_columns = columns ^ df2.columns
        # FIXME: Due to the dynamic nature of attribute analysis, in rare cases
        # it can happen that extra columns do not representing a schema change
        # Example: DF with single row Char that becomes Many2many (extra slots)
        if df1_extra_columns:
            raise ExtraColumnsException(
                "The old schema has extra attributes "
                "not known in the new schema. " + LIB_UPDATE_STR
            )
        if df2_extra_columns:
            raise ExtraColumnsException(
                "The new schema has extra attributes "
                "not known in the old schema. " + LIB_UPDATE_STR
            )

        # Step 3:  Compare normalized dataframes
        changed = df1.loc[rows, columns] != df2.loc[rows, columns]
        df1_changed = df1.loc[rows, columns][changed]
        df2_changed = df2.loc[rows, columns][changed]

        # Example:
        # >>> df1.loc[rows,columns][changed]
        #        boolean       list  set
        # index
        # row1       NaN        NaN  NaN
        # row2       NaN  [1, 2, 3]  NaN
        # >>> df2.loc[rows,columns][changed]
        #        boolean       list  set
        # index
        # row1       NaN        NaN  NaN
        # row2       NaN  [3, 2, 1]  NaN

        # Step 4:  Identify known transitinos and hanlde unkowns
        print(df1_extra_rows)
        print(df2_extra_rows)
        print(df1_changed)
        print(df2_changed)


# 0. define receiving data structures for field spec of both branches
# 1. execute subroutine on two branches
# # 1. checkout branch (recursively)
# # 2. start odoo environment
# # 3. run get_dataframe_of_all_fields
# # 4. stop odoo environment
# 2. return a diff data frame with changed rows
